--------------------------------------------------------------------------------
-- House of Winters - Texture Viewer
--------------------------------------------------------------------------------
-- Exposes the following to the global scope:
--     HoW_TV               (addon namespace)
--     HoW_TV_CT_*          (addon controls)
--
-- Provides the following slash commands:
--     /texview
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Setup module global variable.
--------------------------------------------------------------------------------

HoW_TV = {
    name         = "HoW_TextureViewer",
    displayName  = "Texture Viewer",
    prefix       = "HoW_TV_CT_",
    author       = "Kentarii",
    version      = "1.2.0",
    link         = "http://www.esoui.com/",
    command      = "/txview",
    debug        = false,
}

--------------------------------------------------------------------------------
-- Libraries, defaults and constants.
--------------------------------------------------------------------------------

local lam = LibStub:GetLibrary("LibAddonMenu-1.0")
local json = LibStub:GetLibrary("LibJSON-1.0")

--
-- Module defaults.
--
local defaults = {
    visible = false,
    faded = false,
    alpha = 65,
    anchor = {
        point = TOPLEFT,
        relativePoint = TOPLEFT,
        offsetX = 100,
        offsetY = 100,
    },
    colors = {
        header = "EECA2A",
        -- section = "FF9933",
        text = "FFE3A1",
        value = "F0F0F0",
        selected = "6699FF",
        -- unknown = "E0E0E0",
        -- active = "6699FF",
        -- complete = "208020",
        -- abandoned = "C00000",
    },
    dimension = {
        width = 1000,
        height = 592
        ,
    },
}

--
-- Supported fonts
--
local FONTS = {
   -- "ZoFontBookLetter",
   -- "ZoFontBookNote",
   -- "ZoFontBookPaper",
   -- "ZoFontBookRubbing",
   -- "ZoFontBookScroll",
   -- "ZoFontBookSkin",
   -- "ZoFontBookTablet",
   "ZoFontChat",
   "ZoFontGame",
   "ZoFontGameBold",
   "ZoFontGameLarge",
   "ZoFontGameLargeBold",
   "ZoFontGameLargeBoldShadow",
   "ZoFontGameMedium",
   "ZoFontGameOutline",
   "ZoFontGameShadow",
   "ZoFontGameSmall",
   -- "ZoFontWinT1",
   -- "ZoFontWinT2",
}

--------------------------------------------------------------------------------
-- Addon init and event/command handling functions.
--------------------------------------------------------------------------------


--- Event handler for EVENT_ADD_ON_LOADED.
-- @return void
function HoW_TV.AddOnLoaded(eventCode, addOnName)
    if addOnName ~= HoW_TV.name then return end

    -- Register the slash command handler.
    SLASH_COMMANDS[HoW_TV.command] = HoW_TV.CommandHandler

    -- Initialize addon.
    HoW_TV:Init()

    -- Add addon settings page.
    HoW_TV.Settings()
end

--- Initialize HoW_TV.
-- @return void
function HoW_TV:Init()
    -- Load saved vars.
    HoW_TV.vars = ZO_SavedVars:New("HoW_TextureViewer_vars", 1, "TextureViewer", defaults)
    HoW_TV.textures = json.Deserialize(HoW_TV.texturesJSON)
    HoW_TV.directory = "/esoui/art/achievements"
    HoW_TV.filepath, HoW_TV.file = HoW_TV:GetFirstChild(HoW_TV.directory)

    -- Update control from vars.
    HoW_TV_CT:ClearAnchors()
    HoW_TV:SetAnchor(HoW_TV.vars.anchor)
    HoW_TV.SetAlpha(HoW_TV.vars.alpha)
    HoW_TV:SetDimensions(HoW_TV.vars.dimension)
    HoW_TV:SetFaded(HoW_TV.vars.faded)

    -- Show HoW_TV.
    if HoW_TV.vars.visible then
        HoW_TV:OpenWindow()
    end

    -- Bind event handlers.
    ZO_PreHookHandler(ZO_GameMenu_InGame, "OnShow", function()
        HoW_TV:HideIfVisible()
    end)
    ZO_PreHookHandler(ZO_GameMenu_InGame, "OnHide", function()
        HoW_TV:ShowIfVisible()
    end)
    ZO_PreHookHandler(ZO_InteractWindow, "OnShow", function()
        HoW_TV:HideIfVisible()
    end)
    ZO_PreHookHandler(ZO_InteractWindow, "OnHide", function()
        HoW_TV:ShowIfVisible()
    end)
    ZO_PreHookHandler(ZO_KeybindStripControl, "OnShow", function()
        HoW_TV:HideIfVisible()
    end)
    ZO_PreHookHandler(ZO_KeybindStripControl, "OnHide", function()
        HoW_TV:ShowIfVisible()
    end)
    ZO_PreHookHandler(ZO_MainMenuCategoryBar, "OnShow", function()
        HoW_TV:HideIfVisible()
    end)
    ZO_PreHookHandler(ZO_MainMenuCategoryBar, "OnHide", function()
        HoW_TV:ShowIfVisible()
    end)
end

--- Event handler for OnMoveStop to reposition window per character.
-- @return void
function HoW_TV:OnMoveStop()
    local isValidAnchor, point, relativeTo, relativePoint, offsetX, offsetY = HoW_TV_CT:GetAnchor()
    if (isValidAnchor) then
        HoW_TV.vars.anchor.point = point
        HoW_TV.vars.anchor.relativePoint = relativePoint
        HoW_TV.vars.anchor.offsetX = offsetX
        HoW_TV.vars.anchor.offsetY = offsetY
    end
end

--- Event handler for OnResizeStop to resize window per character.
-- @return void
function HoW_TV:OnResizeStop()
    local width, height = HoW_TV_CT:GetDimensions()
    if width and height then
        HoW_TV.vars.dimension.width = width
        HoW_TV.vars.dimension.height = height
    end
    -- Resize might cause anchor change as well.
    HoW_TV:OnMoveStop()
end

--- Event handler for OnMouseWheel to scroll TextBuffer control.
-- @param delta int scroll up/down step value.
-- @return void
function HoW_TV:OnMouseWheel(_self, delta, ctrl, alt, shift, command)
    local visible = _self:GetNumVisibleLines()
    local lines = _self:GetNumHistoryLines()
    local scrollPosition = _self:GetScrollPosition()

    if ctrl then
        if _self:GetName() == "HoW_TV_CT_DirListTB" and HoW_TV.directory then
            local absdir, reldir
            if delta < 0 then
                absdir, reldir = HoW_TV:GetNextEntry(HoW_TV.directory)
            elseif delta > 0 then
                absdir, reldir = HoW_TV:GetPreviousEntry(HoW_TV.directory)
            end
            if absdir and not (reldir == "__orderedIndex") and not (HoW_TV:endswith(reldir, ".dds")) then
                HoW_TV.directory = absdir
                HoW_TV.filepath, HoW_TV.file = HoW_TV:GetFirstChild(HoW_TV.directory)
                if HoW_TV.filepath then
                    HoW_TV:PrintLocation(HoW_TV.filepath)
                    HoW_TV:PrintTexture(HoW_TV.filepath)
                else
                    HoW_TV:PrintLocation(HoW_TV.directory)
                    HoW_TV_CT_Texture:SetHidden(true)
                end
                HoW_TV:PrintDirectoryList(HoW_TV.textures, HoW_TV.directory, "", "", HoW_TV_CT_DirListTB:GetScrollPosition())
                HoW_TV:PrintFileList(HoW_TV.directory, HoW_TV.file)
            end
        elseif _self:GetName() == "HoW_TV_CT_FileListTB" and HoW_TV.filepath then
            local directory, file
            if delta < 0 then
                directory, file = HoW_TV:GetNextEntry(HoW_TV.filepath)
            elseif delta > 0 then
                directory, file = HoW_TV:GetPreviousEntry(HoW_TV.filepath)
            end
            if file and not (file == "__orderedIndex") and (HoW_TV:endswith(file, ".dds")) then
                HoW_TV.filepath = HoW_TV.directory..'/'..file
                HoW_TV:PrintLocation(HoW_TV.filepath)
                HoW_TV:PrintFileList(HoW_TV.directory, file, HoW_TV_CT_FileListTB:GetScrollPosition())
                HoW_TV:PrintTexture(HoW_TV.filepath)
            end
        end
    else
        if delta < 0 then
            _self:MoveScrollPosition(delta)
        elseif delta > 0 then
            _self:MoveScrollPosition(delta)
        end
    end
end

function HoW_TV:endswith(s, send)
    return #s >= #send and s:find(send, #s-#send+1, true) and true or false
end

function HoW_TV:GetPreviousEntry(path)
    local strpos = string.find(path, "/[^/]*$")
    local parent = zo_strsub(path, 0 , strpos - 1)
    local child = zo_strsub(path, strpos + 1)
    local dirs = HoW_TV:GetFilesForPath(parent)
    local previous = nil
    if dirs then
        for k, v in orderedPairs(dirs) do
            if k == child then
                if not previous then
                    break
                end
                return parent..'/'..previous, previous
            end
            previous = k
        end
    end
    return nil, nil
end

function HoW_TV:GetNextEntry(path)
    local strpos = string.find(path, "/[^/]*$")
    local parent = zo_strsub(path, 0 , strpos - 1)
    local child = zo_strsub(path, strpos + 1)
    local dirs = HoW_TV:GetFilesForPath(parent)
    local next = false
    if dirs then
        for k, v in orderedPairs(dirs) do
            if next then
                return parent..'/'..k, k
            end
            if k == child then
                next = true
            end
        end
    end
    return nil, nil
end

function HoW_TV:GetFirstChild(path)
    local dirs = HoW_TV:GetFilesForPath(path)
    for k, v in orderedPairs(dirs) do
        if v.file then
            return v.path, k
        else
            break
        end
    end
    return nil, nil
end

--- Event handler for OnLinkClicked in TextBuffer control.
-- @param self
-- @param linkData string the link url:     url:SomeLinkTargetOrUrl[SomeLinkText]
-- @param linkText string the link text:    [SomeLinkText]
-- @param button int the mousebutton clicked, 1 = left, 2 = right, 3 = middle
-- @param ctrl bool true if ctrl was pressed during click.
-- @param alt bool true if alt was pressed during click.
-- @param shift bool true if shift was pressed during click.
-- @param command bool true if command (Mac I assume) was pressed during click.
-- @return void
function HoW_TV:OnLinkClicked(_self, linkData, linkText, button, ctrl, alt, shift, command)
    local linkStyle, linkType, arg1, arg2 = zo_strsplit(":", linkData)
    if linkType == "dir" then
        local path = arg1
        local directory = arg2
        if button == 1 then
            HoW_TV.directory = path
            HoW_TV.filepath, HoW_TV.file = HoW_TV:GetFirstChild(HoW_TV.directory)
            if HoW_TV.filepath then
                HoW_TV:PrintLocation(HoW_TV.filepath)
                HoW_TV:PrintTexture(HoW_TV.filepath)
            else
                HoW_TV:PrintLocation(HoW_TV.directory)
                HoW_TV_CT_Texture:SetHidden(true)
            end
            HoW_TV:PrintDirectoryList(HoW_TV.textures, HoW_TV.directory, "", "", HoW_TV_CT_DirListTB:GetScrollPosition())
            HoW_TV:PrintFileList(HoW_TV.directory, HoW_TV.file)
        end
    elseif linkType == "file" then
        local file = arg1
        local name = arg2
        if button == 1 then
            HoW_TV.file = file
            HoW_TV.filepath = HoW_TV.directory..'/'..file
            HoW_TV:PrintLocation(HoW_TV.filepath)
            HoW_TV:PrintFileList(HoW_TV.directory, file, HoW_TV_CT_FileListTB:GetScrollPosition())
            HoW_TV:PrintTexture(HoW_TV.filepath)
        end
    end
end

function HoW_TV:ScrollToFirstLine(_self)
    local visible = _self:GetNumVisibleLines()
    local lines = _self:GetNumHistoryLines()
    _self:SetScrollPosition(lines - (visible + 1))
end

--- Addon command handler.
-- @param text the command line argument (original case)
-- @return void
function HoW_TV.CommandHandler(text)
    local cmd = string.lower(text)

    if cmd == "show" then
        HoW_TV.OpenWindow()
    elseif cmd == "hide" then
        HoW_TV:CloseWindow()
    else
        d("----------------------------------------")
        d("|cEECA2A"..HoW_TV.displayName.."|r |c6699FF"..HoW_TV.version.."|r by |cFF9933"..HoW_TV.author.."|r")
        d("----------------------------------------")
        d("Commands:")
        d(HoW_TV.command.." show")
        d(HoW_TV.command.." hide")
        d("----------------------------------------")
    end
end

--------------------------------------------------------------------------------
-- Setup module control panel section.
--------------------------------------------------------------------------------

--- Addon settings.
-- @return void
function HoW_TV.Settings()
    local controlPanel = HoW_Settings

    -- Create HoW settings page.
    if not controlPanel then
        controlPanel = lam:CreateControlPanel("HoW_Settings", "|cEECA2AHoW Addons|r")
    end

    -- Add module settings.
    lam:AddHeader(controlPanel, HoW_TV.prefix.."SettingsHeader", HoW_TV.displayName)
    lam:AddCheckbox(controlPanel, HoW_TV.prefix.."SettingsVisible", "Visible", "Toggle Texture Viewer visibility", HoW_TV.IsVisible, HoW_TV.SetVisible)
    lam:AddDropdown(controlPanel, HoW_TV.prefix.."SettingsFont", "Font", "Change font family", FONTS, HoW_TV.GetFont, HoW_TV.SetFont)
    lam:AddSlider(controlPanel, HoW_TV.prefix.."SettingsAlpha", "Alpha", "Change background transparency", 0, 100, 1, HoW_TV.GetAlpha, HoW_TV.SetAlpha)
    lam:AddDescription(controlPanel, HoW_TV.prefix.."AboutDescription", "Chat command: |c66FF99"..HoW_TV.command.."|r\n\n|cEECA2A"..HoW_TV.displayName.."|r |c6699FF"..HoW_TV.version.."|r by |cFF9933"..HoW_TV.author.."|r")
end

function HoW_TV.IsVisible()
    return HoW_TV.vars and HoW_TV.vars.visible
end

function HoW_TV.SetVisible(checkbox)
    HoW_TV.vars.visible = checkbox
end

function HoW_TV.GetFont()
    if HoW_TV.vars and HoW_TV.vars.font then
        return HoW_TV.vars.font
    else
        return "ZoFontGame"
    end
end

function HoW_TV.SetFont(font)
    HoW_TV.vars.font = font
    HoW_TV_CT_HeaderLabel:SetFont(font)
    HoW_TV_CT_TextLabel:SetFont(font)
    HoW_TV_CT_DirListTB:SetFont(font)
end

function HoW_TV.GetAlpha()
    if HoW_TV.vars and HoW_TV.vars.alpha then
        return HoW_TV.vars.alpha
    else
        return 65
    end
end

function HoW_TV.SetAlpha(alpha)
    HoW_TV.vars.alpha = alpha
    HoW_TV_CT_Backdrop:SetAlpha(alpha / 100)
    HoW_TV_CT_DirListTB_Backdrop:SetAlpha(alpha / 100)
    HoW_TV_CT_FileListTB_Backdrop:SetAlpha(alpha / 100)
end

--------------------------------------------------------------------------------
-- Show Texture Viewer window.
--------------------------------------------------------------------------------

--- Show TextureViewer window.
-- @param options table with options to show, types/character.
-- @return void
function HoW_TV.OpenWindow()
    -- Remember window state.
    HoW_TV.vars.visible = true

    -- Set font and scale on components
    local font = HoW_TV.GetFont()
    HoW_TV_CT_HeaderLabel:SetFont(font)
    HoW_TV_CT_DirListTB:SetFont(font)
    HoW_TV_CT_FileListTB:SetFont(font)

    HoW_TV:UpdateWindow(true)
    HoW_TV_CT:SetHidden(false)
end

--- Update window with info.
--- Called from xml without args and OpenWindow.
-- @param force trigger update without buffering when opening up window to show content directly.
-- @return void
function HoW_TV:UpdateWindow(force)
    if not force and not BufferReached("HoW_TV.OnUpdate", 1) then return end

    -- Header
    HoW_TV:PrintHeader()

    -- Address
    if HoW_TV.filepath then
        HoW_TV:PrintLocation(HoW_TV.filepath)
    else
        HoW_TV:PrintLocation(HoW_TV.directory)
    end

    -- Directory listing
    HoW_TV:PrintDirectoryList(HoW_TV.textures, HoW_TV.directory, "", "")

    -- File listing
    HoW_TV:PrintFileList(HoW_TV.directory, HoW_TV.file)

    -- Texture view
    if HoW_TV.filepath then
        HoW_TV:PrintTexture(HoW_TV.filepath)
    end
end

--- Print heading with addon name and optionally zonename.
-- @return void
function HoW_TV:PrintHeader()
    HoW_TV_CT_HeaderLabel:SetText(string.format(
        "|c%s%s",
        HoW_TV.vars.colors.header,
        HoW_TV.displayName
    ))
end

--- Update TextLabel with current location.
-- @param location string the current selected location.
-- @return void
function HoW_TV:PrintLocation(path)
    HoW_TV_CT_LocationEB:SetText(path)
end

--- Update TextBuffer with directory listing.
-- @param dirlist table with directory listing.
-- @param location string the current selected location.
-- @param parent the parent directory, used for the links.
-- @param prefix string prefix to add to print out.
-- @param scrollPos optional position to scroll back to.
-- @return void
function HoW_TV:PrintDirectoryList(dirlist, location, parent, prefix, scrollPos)
    if parent == "" then
        HoW_TV_CT_DirListTB:Clear()
    end
    if dirlist and type(dirlist) == "table" then
        for k, v in orderedPairs(dirlist) do
            if not v.file and not (k == "__orderedIndex") then
                local color = HoW_TV.vars.colors.value
                if location == parent..'/'..k then
                    color = HoW_TV.vars.colors.selected
                end

                HoW_TV_CT_DirListTB:AddMessage(string.format(
                    "|c%s%s |c%s|H%d:dir:%s/%s:|h%s|h |c%s(%d)",
                    HoW_TV.vars.colors.text,
                    prefix,
                    color,
                    LINK_STYLE_DEFAULT,
                    parent,
                    k,
                    k,
                    HoW_TV.vars.colors.text,
                    HoW_TV:tablelength(v)
                ), 1, 1, 1, 1);
                if v and type(v) == "table" then
                    HoW_TV:PrintDirectoryList(v, location, parent..'/'..k, prefix.."--")
                end
            end
        end
        if scrollPos then
            HoW_TV_CT_DirListTB:SetScrollPosition(scrollPos)
        else
            HoW_TV:ScrollToFirstLine(HoW_TV_CT_DirListTB)
        end
    end
end

--- Return file(s) for given path, supports 4 nested levels.
-- @param path the given path.
-- @return table with file(s) for given path.
function HoW_TV:GetFilesForPath(path)
    local arg1, arg2, arg3, arg4 = zo_strsplit("/", path)
    if arg1 and HoW_TV.textures[arg1] then
        if arg2 and HoW_TV.textures[arg1][arg2] then
            if arg3 and HoW_TV.textures[arg1][arg2][arg3] then
                if arg4 and HoW_TV.textures[arg1][arg2][arg3][arg4] then
                    return HoW_TV.textures[arg1][arg2][arg3][arg4]
                else
                    return HoW_TV.textures[arg1][arg2][arg3]
                end
            else
                return HoW_TV.textures[arg1][arg2]
            end
        else
            return HoW_TV.textures[arg1]
        end
    else
        return nil
    end
end

function HoW_TV:tablelength(table)
    local count = 0
    for k, v in pairs(table) do
        if not (k == "__orderedIndex") then
            count = count + 1
        end
    end
    return count
end

--- Update TextBuffer with texture files for given directory.
-- @param path the path to print textures for.
-- @param currentFile the currently selected file
-- @param scrollPos optional position to scroll back to.
-- @return void
function HoW_TV:PrintFileList(path, currentFile, scrollPos)
    HoW_TV_CT_FileListTB:Clear()
    local files = HoW_TV:GetFilesForPath(path)
    if files then
        local i = 0
        for k, v in orderedPairs(files) do
            if v.file then
                i = i + 1
                local name, _ = zo_strsplit(".", k)
                local color = HoW_TV.vars.colors.value
                if currentFile and currentFile == k then
                    color = HoW_TV.vars.colors.selected
                end
                if zo_strlen(name) > 50 then
                    name = zo_strsub(name, 0, 48)..'..'
                end
                HoW_TV_CT_FileListTB:AddMessage(string.format(
                    "|c%s%02d |c%s|H%d:file:%s:|h%s|h",
                    HoW_TV.vars.colors.text,
                    i,
                    color,
                    LINK_STYLE_DEFAULT,
                    k,
                    name
                ), 1, 1, 1, 1);
            end
        end
    end
    if scrollPos then
        HoW_TV_CT_FileListTB:SetScrollPosition(scrollPos)
    else
        HoW_TV:ScrollToFirstLine(HoW_TV_CT_FileListTB)
    end
end

--- Update TextBuffer with given texture path
-- @param path the path to print textures for.
-- @return void
function HoW_TV:PrintTexture(path)
    local file = HoW_TV:GetFilesForPath(path)
    -- d(file)
    HoW_TV_CT_Texture:SetTexture(path)
    HoW_TV_CT_Texture:SetHidden(false)
    local width, height = HoW_TV_CT_Texture:GetDimensions()
    if file.width and file.height and file.size then
        HoW_TV_CT_FileInfoL:SetText(string.format(
            "|c%sDimensions: |c%s%d x %d |c%spixels\nFilesize: |c%s%d |c%sBytes|r",
            HoW_TV.vars.colors.text,
            HoW_TV.vars.colors.value,
            file.width,
            file.height,
            HoW_TV.vars.colors.text,
            HoW_TV.vars.colors.value,
            file.size,
            HoW_TV.vars.colors.text
        ))
    elseif width and height then
        HoW_TV_CT_FileInfoL:SetText(string.format(
            "|c%sDimensions: |c%s%d x %d |c%spixels|r",
            HoW_TV.vars.colors.text,
            HoW_TV.vars.colors.value,
            width,
            height,
            HoW_TV.vars.colors.text
        ))
    end
    -- Print out any empty textures when in debug mode.
    if HoW_TV.debug then
        if (width == 0) or (height == 0) then
            d(string.format("%s (%dx%d)", path, width, height))
        end
    end
end

--------------------------------------------------------------------------------
-- Update xml control functions.
--------------------------------------------------------------------------------

--- Set window anchor position.
-- @param anchor table with point, relativePoint, offsetX and offsetY.
-- @return void
function HoW_TV:SetAnchor(anchor)
    HoW_TV_CT:SetAnchor(anchor.point, GuiRoot, anchor.relativePoint, anchor.offsetX, anchor.offsetY)
end

--- Set window size.
-- @param dimension table with width and height
-- @return void
function HoW_TV:SetDimensions(dimension)
    HoW_TV_CT:SetDimensions(dimension.width, dimension.height)
end

--- Set window faded.
-- @param faded boolean faded state
-- @return void
function HoW_TV:SetFaded(faded)
    HoW_TV_CT_Backdrop:SetHidden(faded)
end


--- Fade window backdrop.
-- @return void
function HoW_TV:FadeWindow()
    -- Remember window faded state.
    HoW_TV.vars.faded = not HoW_TV_CT_Backdrop:IsHidden()
    HoW_TV:SetFaded(HoW_TV.vars.faded)
end

--- Refresh window with current zone.
-- @return void
function HoW_TV:RefreshWindow()
    HoW_TV:UpdateWindow(true)
end

--- Close the Texture Viewer window.
-- @return void
function HoW_TV:CloseWindow()
    -- Remember window state.
    HoW_TV.vars.visible = false
    HoW_TV_CT:SetHidden(true)
end

--- Hide the window if visible and screen state changed to some dialogue/menu interface.
-- @return void
function HoW_TV:HideIfVisible()
    if HoW_TV.vars.visible then
        HoW_TV_CT:SetHidden(true)
    end
end

--- Show the window if visible and screen state returned to normal.
-- @return void
function HoW_TV:ShowIfVisible()
    if HoW_TV.vars.visible then
        HoW_TV_CT:SetHidden(false)
    end
end

--------------------------------------------------------------------------------

-- Initialize addon.
EVENT_MANAGER:RegisterForEvent(HoW_TV.name, EVENT_ADD_ON_LOADED, HoW_TV.AddOnLoaded)
